////////////////////////////////////////////////////////////////////////
/// \file  SingleGen_plugin.cc
/// \brief Generator for cosmic-rays
///
/// Module designed to produce a set list of particles for a MC event
///
/// \version $Id: SingleGen.cxx,v 1.4 2010/03/29 09:54:01 brebel Exp $
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////




// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h" 
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

#include "TFile.h"
#include "TTree.h"
#include "TChain.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TVector3.h"
#include "TDatabasePDG.h"
#include "TRandom3.h"
#include "TSystem.h"
#include "TSystemDirectory.h"
#include "TSystemFile.h"

#include <sstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <dirent.h>
#include <time.h>
#include <stdio.h>
#include <cstdlib>
#include <stdlib.h>
#include <math.h>
#include <vector>
#include <string>
#include "ifdh.h"

// nutools includes
#include "nusimdata/SimulationBase/MCTruth.h"
#include "nutools/RandomUtils/NuRandomService.h"

// lar includes
#include "larcore/Geometry/Geometry.h"
#include "larcoreobj/SummaryData/RunData.h"
#include "larcoreobj/SummaryData/POTSummary.h"

#include "CLHEP/Random/RandFlat.h"


namespace simb { class MCTruth; }

namespace t962 {

  /// module to produce single or multiple specified particles in the detector
  class TGMuon : public art::EDProducer {

  public:
    explicit TGMuon(fhicl::ParameterSet const& pset);
    virtual ~TGMuon();

    // This is called for each event.
    void produce(art::Event& evt);
    void beginRun(art::Run& run);
    void endSubRun(art::SubRun& sr);
    void reconfigure(fhicl::ParameterSet const& pset) {};

  private:

    void Sample(simb::MCTruth &mct);        
    
    //char fBuffer[128]; // unused
    //TDirectory* fSaveDir ; // unused
    TString fpwd;
    int frun;
    int fsubRun;
    int fsnarl;
    int fntrkstp;
    float ftrkIndex;
    float ftrtgtd;
    float ftrkVtxX;
    float ftrkVtxY;
    float ftrkVtxZ;
    float ftrkdcosx;
    float ftrkdcosy;
    float ftrkdcosz; 
    float ftrkmom;
    float ftrkqp;
    float totalpot;

    int fTIDOffset;
    bool fanu;
    std::vector<TString> fanuFileVector;

    std::string              fFluxSearchPaths;   ///< colon separated set of path stems
    std::vector<std::string> fFluxFilePatterns;  ///< wildcard patterns files containing histograms or ntuples, or txt
    std::vector<std::string> fSelectedFluxFiles; ///< flux files selected after wildcard expansion and subset selection
    int                      fMaxFluxFileMB;     ///< maximum size of flux files (MB)

    TChain *chain;
    Long64_t nentries;

    ifdh_ns::ifdh* fIFDH;
    void OpenInputFile();

    CLHEP::HepRandomEngine& fEngine; ///< random engine

  };
  

  //____________________________________________________________________________
  TGMuon::TGMuon(fhicl::ParameterSet const& pset)
    : art::EDProducer{pset}
    ,fTIDOffset(100)
    ,fanu(true)
    ,fEngine(art::ServiceHandle<rndm::NuRandomService>{}->createEngine(*this, pset, "Seed"))
  {
    this->reconfigure(pset);
    produces< std::vector<simb::MCTruth> >();
    produces< sumdata::RunData, art::InRun >();
    produces< sumdata::POTSummary, art::InSubRun >();
    totalpot=0.;    
    fanu       = pset.get< bool >("AntiNeutrinos", false);
    fFluxSearchPaths = pset.get< std::string >("FluxSearchPaths");
    fFluxFilePatterns = pset.get< std::vector<std::string> >("FluxFiles");
    fMaxFluxFileMB = pset.get< int >("MaxFluxFileMB",500);
    fIFDH = 0;
    chain = 0;
    nentries = 0;

    OpenInputFile();
  }

  //____________________________________________________________________________
  TGMuon::~TGMuon()
  {
    if (fIFDH){
      auto ffitr = fSelectedFluxFiles.begin();
      for ( ; ffitr != fSelectedFluxFiles.end(); ++ffitr ) {
        std::string ff = *ffitr;
        if ( ff.find("/var/tmp") == 0 ) {
          mf::LogDebug("TGMuon") << "delete " << ff;
          fIFDH->rm(ff);
        }
      }
      delete fIFDH;
      fIFDH = 0;
    }
    if (chain){
      delete chain;
      chain = 0;
    }
  }

  //____________________________________________________________________________

  //____________________________________________________________________________
  void TGMuon::beginRun(art::Run& run)
  {
    // grab the geometry object to see what geometry we are using
    art::ServiceHandle<geo::Geometry> geo;
    //geo::DetId_t detid = geo->DetId();
    std::unique_ptr<sumdata::RunData> runcol(new sumdata::RunData("argoneut"));
    run.put(std::move(runcol));

    return;

  }

  void TGMuon::endSubRun(art::SubRun& sr)
  {
    std::unique_ptr<sumdata::POTSummary> p(new sumdata::POTSummary());    
    p->totpot = totalpot;
    p->totgoodpot = totalpot;    
    sr.put(std::move(p));    
    return;
  }

  //____________________________________________________________________________
  void TGMuon::produce(art::Event& evt)
  {
    ///unique_ptr allows ownership to be transferred to the art::Event after the put statement
    std::unique_ptr< std::vector<simb::MCTruth> > truthcol(new std::vector<simb::MCTruth>);
    simb::MCTruth truth;
    truth.SetOrigin(simb::kSingleParticle);
    Sample(truth);
    truthcol->push_back(truth);
    evt.put(std::move(truthcol));

    return;
  }

  //____________________________________________________________________________
  void TGMuon::Sample(simb::MCTruth &mct) 
  {
    double x_offset=117.4; 
    double y_offset=19.3;
    double z_offset=150.; //force particle to be well upstream of ArgoNeuT

    CLHEP::RandFlat flat(fEngine);
    int snarl=0;
    int run=0;
    int subRun=0;
    int flag=0;
    int flag2=0;
    int r = 0;
    while(flag2==0){      
      r=flat.fireInt(0,nentries-1);
      chain->GetEntry(r);
      run=frun;
      subRun=fsubRun;
      snarl=fsnarl;
      flag = 0;
      //count POT even if the particle is not passed to Geant
      totalpot+=ftrtgtd*pow(10,12);
          
      for(int i=int(r-ftrkIndex);i<nentries;i++){
        if (i<0) continue;
        chain->GetEntry(i);

        if(frun==run && fsubRun==subRun && fsnarl==snarl)
          flag=1;
        
        if(flag==1 && fsnarl!=snarl)
     	  break;

        if (!flag) continue;

        TVector3 x;
        x[0]=100.*ftrkVtxX - x_offset;
        x[1]=100.*ftrkVtxY + y_offset;
        x[2]=100.*ftrkVtxZ - z_offset;
        //don't bother passing the particle to geant if it's way outside the argoneut detector window   
      	if(100.0*ftrkVtxZ > 20. || fabs(x[0]+23.5)>100. || fabs(x[1])>100.)
          continue;
        
      	// Choose momentum
      	double p = 0.0;
      	double m = 0.0;
      	p=ftrkmom;
        
      	if(!p)
          continue;
        
        if (p>120.) p = 120.; //protect against craze momentum
        
      	int pdg=0;
      	
      	if(ftrkqp>0)
          pdg=-13;
      	else
          pdg=13;
        
      	TParticlePDG* pdgp = TDatabasePDG::Instance()->GetParticle(pdg);
      	if (pdgp) m = pdgp->Mass();
        
      	TLorentzVector pos(x[0], x[1], x[2], 0.0);
      	TLorentzVector pvec(ftrkdcosx*p,ftrkdcosy*p,ftrkdcosz*p,sqrt(p*p+m*m));
        
      	int trackid = -1*(i+1+fTIDOffset); // set track id to -i as these are all primary particles and have id <= 0
      	std::string primary("primary");
        
      	simb::MCParticle part(trackid, pdg, primary);
      	part.AddTrajectoryPoint(pos, pvec);       
      	mct.Add(part);
	flag2=1;      
      }//for loop
    }//while flag2 loop
    
    return;
  }

  //____________________________________________________________________________
  void TGMuon::OpenInputFile(){	

    CLHEP::RandFlat flat(fEngine);

    // Setup ifdh object
    if (!fIFDH) {	
      fIFDH = new ifdh_ns::ifdh;
    }

    std::string spaths = fFluxSearchPaths;

    const char* ifdh_debug_env = std::getenv("IFDH_DEBUG_LEVEL");
    if ( ifdh_debug_env ) {
      mf::LogInfo("TGMuon") << "IFDH_DEBUG_LEVEL: " << ifdh_debug_env<<"\n";
      fIFDH->set_debug(ifdh_debug_env);
    }

    // filenames + size 
    std::vector<std::pair<std::string,long>> 
      partiallist, fulllist, selectedlist, locallist;
  
    mf::LogVerbatim("TGMuon")<< "search paths: " << spaths;
    // loop over possible patterns
    // IFDH handles various stems but done a list of globs
    size_t ipatt=0;
    auto uitr = fFluxFilePatterns.begin();
    for ( ; uitr != fFluxFilePatterns.end(); ++uitr, ++ipatt ) {
      std::string userpattern = *uitr;
      mf::LogVerbatim("TGMuon") << "\npattern [" << std::setw(3) << ipatt << "] " << userpattern;

      partiallist = fIFDH->findMatchingFiles(spaths,userpattern);
      fulllist.insert(fulllist.end(),partiallist.begin(),partiallist.end());

      // make a complete list ...
      mf::LogDebug("TGMuon") << " found " << partiallist.size() << " files";
      for (auto pitr = partiallist.begin(); pitr != partiallist.end(); ++pitr) {
        mf::LogDebug("TGMuon")  << std::setw(10) << pitr->second << " " << pitr->first;
      }
      partiallist.clear();      
    }  // loop over user patterns 
    
    size_t nfiles = fulllist.size();
    
    mf::LogVerbatim("TGMuon") << "list of " << nfiles << " will be randomized and pared down to " << fMaxFluxFileMB << " MB";
    
    double* order = new double[nfiles];
    int* indices  = new int[nfiles];
    
    flat.fireArray(nfiles, order);
    
    TMath::Sort((int)nfiles,order,indices,false);
    long long int sumBytes = 0; // accumulated size in bytes
    long long int maxBytes = (long long int)fMaxFluxFileMB * 1024ll * 1024ll;

    for (size_t i=0; i<nfiles; ++i) {
      int indx = indices[i];
      bool keep = true;
        
      auto p = fulllist[indx]; // this pair <name,size>
      sumBytes += p.second; 
      // skip those that would push sum above total
      // but always accept at least one (the first)
      if ( sumBytes > maxBytes && i != 0 ) keep = false;
      
      mf::LogVerbatim("TGMuon") << std::setw(3) << i << "] "
                                << "=> [" << std::setw(3) << indx << "] " 
                                << ((keep)?"keep":"SKIP") << " " 
                                << std::setw(6) << (sumBytes/(1024ll*1024ll)) << " MB "
                                << p.first;
      
      if ( keep ) selectedlist.push_back(p);
      else break;  // <voice name=Scotty> Captain, she can't take any more</voice>
      
    }
    delete [] order;
    delete [] indices;

    chain = new TChain("minitree");

    // have a selected list of remote files
    // get paths to local copies
    locallist = fIFDH->fetchSharedFiles(selectedlist, "");

    mf::LogVerbatim("TGMuon") << "final list of files:";
    size_t i=0;
    for (auto litr = locallist.begin(); litr != locallist.end(); ++litr, ++i) {
      fSelectedFluxFiles.push_back(litr->first);
      mf::LogVerbatim("TGMuon") << "\t[" << std::setw(3) << i << "]\t" << litr->first;
      chain->Add((litr->first).c_str());
    }

    nfiles = fSelectedFluxFiles.size();
    if ( nfiles == 0 ) {
      mf::LogError("TGMuon") 
        << "must resolve to at least one file" 
        << "\n  none were found user pattern(s): " 
        << "\n  using FW_SEARCH_PATH of: "
        << spaths;
      
      throw cet::exception("NoMINOSFiles")
        << "no MINOS files found for: " << spaths;
    }
    nentries = chain->GetEntries();
    mf::LogVerbatim("TGMuon") <<"Number of entries in the tchain = "<<nentries;
    if (!nentries){
      throw cet::exception("NoMINOSSnarls")
        << "no MINOS snarls found in the tchain";
    }
    chain->SetBranchAddress("run",&frun);  
    chain->SetBranchAddress("subRun",&fsubRun);            
    chain->SetBranchAddress("trkIndex",&ftrkIndex);
    chain->SetBranchAddress("ntrkstp",&fntrkstp);
    chain->SetBranchAddress("snarl",&fsnarl);
    chain->SetBranchAddress("trtgtd",&ftrtgtd);
    chain->SetBranchAddress("trkVtxX",&ftrkVtxX);
    chain->SetBranchAddress("trkVtxY",&ftrkVtxY);
    chain->SetBranchAddress("trkVtxZ",&ftrkVtxZ);
    chain->SetBranchAddress("trkdcosx",&ftrkdcosx);
    chain->SetBranchAddress("trkdcosy",&ftrkdcosy);
    chain->SetBranchAddress("trkdcosz",&ftrkdcosz);
    chain->SetBranchAddress("trkmom",&ftrkmom);
    chain->SetBranchAddress("trkqp",&ftrkqp);            

  }

  DEFINE_ART_MODULE(TGMuon)

}//end namespace t962
////////////////////////////////////////////////////////////////////////
